package chase.s2.stat;

import org.csdgn.utils.VDA;

import chase.s2.Tools;

/**
 * This is a visit count stat buffer. However, it is such a buffer where it may
 * have any number of segmentation or dimensionality.
 * 
 * For those uninformed, this is part of a statistical data mining system.
 * 
 * @author Chase
 */
public class StatBuffer {
	protected static final int scanWidth = 1;
	
	/** Handing the VDA manually is dangerous, and could wipe out all your data */
	private VDA<double[]> _binSet;
	public int _binTotal;
	public int[] _dataIndex;
	/** This is fine, since java can handle jagged arrays */
	public double[][] _splits;
	
	double decayRate = -1;
	
	/**
	 * Constructs a new buffer
	 * @param splits The data defining the data splits, usually a jagged array
	 * @param dataIndex the index of the raw stats each split is on
	 * @param binTotal total number of indexes for this buffer
	 */
	public StatBuffer(double[][] splits, int[] dataIndex, int binTotal) {
		/* TODO: splits and dataIndex lengths don't equal throw exception */
		_binTotal = binTotal;
		_dataIndex = dataIndex;
		_splits = splits;
		
		int[] dimensions = new int[Math.max(1, splits.length)];
		dimensions[0] = 1;
		for(int i=0; i<splits.length; ++i)
			dimensions[i] = splits[i].length + 1;
		
		_binSet = new VDA<double[]>(dimensions);
	}
	
	public void initialize(int[] pos) {
		double[] d = _binSet.get(pos);
		if(d == null) {
			_binSet.set(new double[_binTotal], pos);
		}
	}
	
	/**
	 * Constructs a new stat buffer, using the lazy SplitSet
	 * @param ss SplitSet to use
	 * @param binTotal total number of indexes for this buffer
	 */
	public StatBuffer(SplitSet ss, int binTotal) {
		this(ss.getSplitSet(),ss.getIndexSet(),binTotal);
	}
	
	/**
	 * This creates a flat, single dimension stat buffer.
	 * @param binTotal total number of indexes for this buffer
	 */
	public StatBuffer(int binTotal) {
		this(new double[0][0], new int[0], binTotal);
	}
	
	/**
	 * This method splits the raw data as defined by the data index
	 * and split information. It outputs a data array which can be
	 * fed to the VDA to get the desired bin set. 
	 * @param data raw data
	 * @return data indexes
	 */
	public int[] convertData(StatData data) {
		int[] indexes = new int[Math.max(1, _dataIndex.length)];
		if(_splits.length == 0) return indexes;
		for(int i = 0; i < indexes.length; ++i) {
			int index = 0;
			for(;index < _splits[i].length; ++index) {
				if(data._data[_dataIndex[i]] <  _splits[i][index]) break;
			}
			indexes[i] = index;
		}
		return indexes;
	}
	
	/**
	 * Gets the individual stat buffer for the give data.
	 */
	public double[] getStatBin(StatData data) {
		int[] indexes = convertData(data);
		initialize(indexes);
		return _binSet.get(indexes);
	}
	
	/**
	 * Sets the decay halflife of this buffer
	 * anything < 0 is infinite.
	 * @param hlife
	 */
	public void setHalflife(double hlife) {
		if(hlife < 0) decayRate = -1;
		else decayRate = Tools.computeDecayRate(hlife);
	}
	
	/**
	 * Updates this stat buffer with the given raw data
	 */
	public void update(StatData data, double weight) {
		double[] bins = getStatBin(data);
		double index = GFToIndex(data._gf, _binTotal);
		if(decayRate > 0) {
			updateBinRolling(bins, index, weight, decayRate);
		} else {
			updateBin(bins, index, weight);
		}
	}
	
	/**
	 * Gets the best bin from the raw data
	 */
	public int getBestIndex(StatData data) {
		double[] bins = getStatBin(data);
		return getBestIndex(bins, scanWidth);
	}
	
	/**
	 * Gets the best guessfactor from the raw data
	 */
	public double getBestGF(StatData data) {
		return indexToGF(getBestIndex(data),_binTotal);
	}
	
	/**
	 * TODO: Replace this with some kind of externalized update method.
	 * That way it is easier to replace with rolling and different distributions
	 */
	public static final void updateBin(double[] bin, double index, double weight) {
		for(int i=0; i<bin.length; ++i) {
			double amount = Math.abs(i - index);
			if(amount == 0) bin[i] += weight;
			if(amount > 0 && amount < 1)
				bin[i] += amount * weight;
		}
	}
	
	public static final void updateBinRolling(double[] bin, double index, double weight, double decay) {
		for(int i=0; i<bin.length; ++i) {
			double amount = Math.abs(i - index);
			if(amount == 0) amount = weight;
			else if(amount > 0 && amount < 1)
				amount = amount * weight;
			else 
				amount = 0;
			
			bin[i] = Tools.timeWeightedAverage(amount, bin[i], 1, decay);
		}
	}
	
	/**
	 * Static method used to find the best index. Width is number
	 * of extra bins to each side to test.
	 */
	public static final int getBestIndex(double[] bin, int width) {
		double[] score = getBinScores(bin,width);
		int bestIndex = (bin.length-1)/2; /* TODO: This could be problematic */
		for(int i=0; i<bin.length; ++i)
			if(score[i] > score[bestIndex])
				bestIndex = i;
		return bestIndex;
	}
	
	/**
	 * Gets the scores for each bin in the given stat bin.
	 */
	protected static final double[] getBinScores(double[] bin, int width) {
		double[] scores = new double[bin.length];
		for(int i=0; i<bin.length; ++i) {
			double score = 0;
			//It is a little complicated in there, but things will
			//   never go out of bounds
			for(int j=-width; j<width+1; ++j) {
				int abs = Math.abs(j) + 1;
				if(i+j < 0 || i+j >= bin.length) {
					if(i == 0 || i == bin.length - 1) {
						score += bin[i - j] / abs;
					} else {
						int offset = (j < 0) ? -3 : 3;
						score += bin[i - j + offset] / abs;
					}
				} else {
					score += bin[i + j] / abs;
				}
			}
			scores[i] = score;
		}
		return scores;
	}
	
	/**
	 * Converts a bin index to a guessfactor based on total number of bins.
	 */
	public static final double indexToGF(int index, int bins) {
		index = Math.min(Math.max(0,index),bins - 1);
		double half = (bins-1.0)/2.0;
		return (index / half) - 1.0;
	}
	
	/**
	 * Converts a guessfactor to a bin index, based on number of bins. This
	 * returns a double, incase there is any need to know how much it over
	 * hangs each bin. Cast to an int for the raw information.
	 */
	public static final double GFToIndex(double gf, int bins) {
		gf = Math.min(Math.max(-1.0,gf),1.0) + 1.0;
		double half = (bins-1.0)/2.0;
		return gf * half;
	}
}