package gh.micro;
import robocode.*;
import robocode.util.Utils;
import java.awt.Color;
import java.awt.geom.*;
/**
 * GrubbmThree - a robot by Gert Heijenk
 *
 * Ramming. Tracks and rams the nearest robot it sees
 *
 * Revision information:
 * v0.1 20041124 Initial version
 * v0.2 20041127 Solved sharp-turn bug, removed 10-tick wait at startup
 * v0.3 20041202 Changed power-distance ratio and made sharp-turn less sharp
 * v0.4 20041218 Use a mix between CT and HOT gun
 * v0.5 20041219 v0.3 with sharp sharp-turn and fire also when angle between 10-20 degrees
 * v0.6 20050127 improved radar, improved rambonus and removed sharp-turn, it slows down the bot too much
 * v0.7 20050130 improved and refactored CT-gun
 * v0.8 20050316 shoot at the edge of the robot, not at the middle
 * v0.9 20060216 shrinking, position aiming-point always in field, less energy-preservation
 */
public class GrubbmThree extends AdvancedRobot
{
    static final double WALL_MARGIN = 17;

	// scanning and locking variables
	static double	oldHeading;		// the previous heading of the selected opponent
	static double	fieldWidth;		// the width of the battlefield
	static double	fieldHeight;	// the heigth of the battlefield
	static Rectangle2D.Double fireField;	// the shootable battlefield

	static Point2D.Double	myNextPos = new Point2D.Double();	// my next position
		
	/**
	 * run: Grubbm_Three's default behavior
	 */
	public void run() {

		// Give the robot an appealing look
		setColors( Color.red, Color.yellow, Color.red);

		// First set the statistics
		fieldWidth = getBattleFieldWidth() - WALL_MARGIN;
		fieldHeight = getBattleFieldHeight() - WALL_MARGIN;
        fireField = new Rectangle2D.Double( WALL_MARGIN, WALL_MARGIN, fieldWidth - WALL_MARGIN, fieldHeight - WALL_MARGIN);

		// Let gun and radar move independently
		setAdjustGunForRobotTurn( true);
		setAdjustRadarForGunTurn( true);

		// mainloop
		do {
			turnRadarRight( Double.NEGATIVE_INFINITY);
		} while(true);
	}

	/**
	 * onScannedRobot: What to do when you see another robot
	 */
	public void onScannedRobot(ScannedRobotEvent e) {

		double power;		// the power of the bullet
		double bearingFromGun;
		double turnAngle;
		double tmpbearing = getHeadingRadians() + e.getBearingRadians();

		// Lock target with radar
		setTurnRadarRightRadians( 2.2 * Utils.normalRelativeAngle( tmpbearing - getRadarHeadingRadians()));

		// set the power of the bullet, 
		power = Math.min( getEnergy() - 0.1, (e.getDistance() > 200 ? 1.9 : 3.0));
//		power = Math.max( 0.0, Math.min( getEnergy() - 1.3, Math.min(((e.getEnergy() - 1) / 4), power)));
//		power = Math.min( 450 / e.getDistance(), Math.min( getEnergy() - 0.1, 3.0));

		// perform iterative circular targeting
		Point2D.Double point = new Point2D.Double();
		long deltahittime;
		double head, chead, speed, bspeed;
		double tmpx, tmpy;
		double dx, dy, edist;

		myNextPos.setLocation( getX() + Math.sin( getHeadingRadians()) * getVelocity(), getY() + Math.cos( getHeadingRadians()) * getVelocity());	// Next position (approx)

		// perform an iteration to find a reasonable accurate expected position
		tmpx = getX() + Math.sin( tmpbearing) * e.getDistance();
		tmpy = getY() + Math.cos( tmpbearing) * e.getDistance();
		head = e.getHeadingRadians();
		chead = head - oldHeading;
		oldHeading = head;
		speed = e.getVelocity();
		point.setLocation( tmpx, tmpy);
		deltahittime = -1;				// necessary to shoot to enemies NEXT position
		do {
			tmpx += Math.sin( head) * speed;
			tmpy +=	Math.cos( head) * speed;
			head += chead;
			deltahittime++;
			// some stupid precise prediction
//			dx = Math.abs(tmpx - myNextPos.getX());
//			dy = Math.abs(tmpy - myNextPos.getY());
//			edist = ((dx > dy) ? (dy * 18) / dx : (dx * 18) / dy);
//			edist = Math.sqrt( 324 + edist * edist);
			// if position not in field, adapt
			if ((!fireField.contains( tmpx, tmpy)) && (deltahittime > 0)) {
				point.setLocation( warpPoint( tmpx, tmpy));
//				bspeed = (point.distance( myNextPos) - edist) / deltahittime;
				bspeed = (point.distance( myNextPos) - 18) / deltahittime;
				if (bspeed < 19.7) power = (20 - bspeed) / 3.0;
//				else if ((point.distance( myNextPos) / 19.7) > (deltahittime + 5)) power = 0.0;
				else power = 0.1;
				break;
			}
			point.setLocation( tmpx, tmpy);
		} while ( (int)Math.round( (point.distance( myNextPos) - 18) / (20 - (3 * power))) > deltahittime);
//		} while ( (int)Math.round( (point.distance( myNextPos) - edist) / (20 - (3 * power))) > deltahittime);

		point.setLocation( warpPoint( tmpx, tmpy));
		// Turn gun
		tmpbearing = ((Math.PI / 2) - Math.atan2( point.y - myNextPos.getY(), point.x - myNextPos.getX()));
		bearingFromGun = Utils.normalRelativeAngle( tmpbearing - getGunHeadingRadians());
		setTurnGunRightRadians( bearingFromGun);
		// If the gun is too hot or not aligned, don't bother trying to fire.
		if ((e.getDistance() <= 350) && (power > 0.0) && (getGunHeat() == 0.0) && (Math.abs( bearingFromGun) < 0.35))
		{
			// Only fire the gun when less than 20 degrees to turn.
			setFire( power);
		}
		// Move towards target
		turnAngle = Utils.normalRelativeAngle( tmpbearing - getHeadingRadians());
		setTurnRightRadians( Math.atan( Math.tan( turnAngle)));
		// sharp turn when necessary
//		setMaxVelocity( Math.abs(myTurnAngle) > 0.785 ? 4.0 : 8.0); // 0.785 -> 45 degrees
//		setMaxVelocity( Math.abs(myTurnAngle) > 0.523 ? 6.0 : 8.0); // 0.523 -> 30 degrees
		setAhead( Double.POSITIVE_INFINITY * Math.cos( turnAngle));

		// Helps in melee to lock onto nearest target.
//		clearAllEvents();
	}

	/**
	 * warpPoint: Put point into field with a margin of 2 * WALL_MARGIN (34)
	 */
	public Point2D.Double warpPoint( double x, double y)
	{
		Point2D.Double pos = new Point2D.Double();

		pos.x = Math.min( fieldWidth  - WALL_MARGIN, Math.max( 2 * WALL_MARGIN, x));
		pos.y = Math.min( fieldHeight - WALL_MARGIN, Math.max( 2 * WALL_MARGIN, y));
		return pos;
	}

	/**
	 * onWin: Show my private victory dance
	 */
	public void onWin(WinEvent e)
	{
		//Victory dance	
		setTurnGunRight( Double.POSITIVE_INFINITY);
		setTurnRadarLeft( Double.POSITIVE_INFINITY);
		setTurnRight(10);
		ahead(10);
		waitFor(new RadarTurnCompleteCondition(this));
	}
}
